"""NTUserFolder: a user folder that authenticates from a NT domain

Notes:

- NTUserFolder est insensible  la casse, mme pour les utilisateur Zope

"""

import time, string, re
from string import lower, strip
from fnmatch import fnmatch
from types import TupleType
from Globals import HTMLFile, PersistentMapping, default__class_init__
from Acquisition import Acquired
from App.Dialogs import MessageDialog
from AccessControl.User import User, UserFolder, domainSpecMatch
try: from AccessControl.User import Super # pour la migration
except: pass
from unicode_support import make_string
from local_ntdom import get_local_ntdom, get_usernames, get_userinfo, authenticate_user

_re_ntusername = re.compile(r'([^\\]+)\\([^\\]+)$')

def _isTop(self):
    """Return true if self is the toplevel"""
    return self.aq_inner.aq_parent.isTopLevelPrincipiaApplicationObject

class NTUserFolder(UserFolder):
    """NT User Folder"""

    _isAjcNTUserFolder = 1

    meta_type = 'NT User Folder'
    id = 'acl_users'
    title = 'NT User Folder'

    _first_options = (
        {'label': 'Detailed view', 'action': 'manage_viewForm'},
        )
    _other_options = (
        {'label': 'Properties', 'action': 'manage_editPropertiesForm'},
        )
    
    manage_options = _first_options + \
                     UserFolder.manage_options[:1] + \
                     _other_options + \
                     UserFolder.manage_options[1:]

    __ac_permissions__=(
        ('Manage users',
        ('isDisabled', 'isChanged',
         'getUserNames', 'getUsers', 'getUser',
         'manage_viewForm',
         'manage_editUserForm', 'manage_editUser',
         'manage_users',
         'manage_editPropertiesForm', 'manage_editProperties',
         )),
        )

    def _normalize(self, login):
        """normalize a login name.
        Currently, we make sure it is a string then it is made lowercase
        """
        return lower(make_string(login))
    
    def _parse_ntusername(self, login, default_ntdom=None, strict=0):
        """return a tuple (name, ntdom, ntname) from a login name '(ntdom\\)?ntname'
        login has already been normalized
        If strict is true, we require the 'ntdom' part
        If login is not a valid nt username, return (login, None, login)
        """
        if default_ntdom is None:
            default_ntdom = self.default_ntdom
        
        mo = _re_ntusername.match(login)
        if mo:
            ntdom, ntname = mo.groups()
            name = ntdom + '\\' + ntname
        else:
            if strict or '\\' in login:
                ntdom = None
                ntname = name = login
            else:
                ntdom = default_ntdom
                ntname = login
                name = ntdom + '\\' + ntname

        return (name, ntdom, ntname)

    def _get_userinfo(self, ntname, ntdom):
        """try to get informations about the NT User 'ntdom\\ntname'
        return None if user is not found
        """
        if ntdom is None or ntdom != self.local_ntdom:
            return None
        return get_userinfo(ntname, ntdom)
    
    def __init__(self, ntdom='', acquire=0, roles=None, domains=None, isTop=1):
        UserFolder.__init__(self)

        self._version = 1
        self.isTop = isTop
        self.local_ntdom = get_local_ntdom()
        self.default_ntdom = lower(strip(ntdom)) or self.local_ntdom
        self.default_acquire = acquire and 1 or 0
        if roles is None: roles = []
        self.default_roles = roles
        if domains is None: domains = []
        self.default_domains = domains

    def __migrate(self):
        """migrate pre-0.1.0 user folders"""
        # check version
        if hasattr(self, '_version') and self._version >= 1:
            # already migrated
            return
        # delete unused variables
        if hasattr(self, 'super'): del self.super
        if hasattr(self, '_super'): del self._super
        if hasattr(self, 'isTop'): del self.isTop
        # migrate roles settings
        oldData = self.data
        newData = self.data = PersistentMapping()
        for name, (local, roles, domains, acquire) in oldData.items():
            name, ntdom, ntname = self._parse_ntusername(self._normalize(name), strict=1)
            if ntdom is None:
                # this is not a valid NT Username. drop it
                print "WARNING: ignoring invalid name %s" % name
                continue
            info = self._get_userinfo(ntname, ntdom) or \
                   {'full_name': '', 'comment': ''}
            user = NTUser(name, None, roles, domains, ntdom, ntname,
                          info['full_name'], info['comment'], acquire,
                          ntuser=1)
            newData[name] = user
        # default values
        self.default_roles = []
        self.default_domains = []
        # done
        self._version = 1

    def isDisabled(self, user):
        """Return true if user is disabled"""
        return user.acquire and \
               (self.isTop or user.roles is self.default_roles)

    def isChanged(self, user):
        """Return true if user settings are changed from the defaults"""
        return user.roles is not self.default_roles or \
               user.domains is not self.default_domains or \
               user.acquire is not self.default_acquire

    def getUserNames(self, exclude_disabled=1, pattern=None):
        """Return a list of usernames
        If exclude_disabled is false, a user name is returned even if the user is
        disabled/meant to be acquired

        If pattern is specified, only usernames matching the pattern according
        to the rules used by the unix shell are returned. i.e. *, ? and
        characters ranges expressed with [] are matched.
        """
        if pattern:
            if '*' not in pattern and '?' not in pattern and \
               '[' not in pattern and ']' not in pattern:
                # if no globbing characters, assume *pattern*
                pattern = '*' + pattern + '*'
            if pattern[0] == '\\':
                # \pattern is a shorthand for [any domain]\pattern
                pattern = '*' + pattern
        names = {}
        ignored = []
        ignore = ignored.append
        for name, user in self.data.items():
            if exclude_disabled and self.isDisabled(user):
                # do not list acquired/disabled users
                ignore(name)
                continue
            if pattern and not fnmatch(name, pattern):
                continue
            names[name] = 1
        if not (exclude_disabled and self.default_acquire):
            for name in get_usernames(self.local_ntdom):
                name = lower(name)
                if pattern and not fnmatch(name, pattern):
                    continue
                if name not in ignored:
                    names[name] = 1
        names = names.keys()
        names.sort()
        return names

    def getUsers(self, exclude_disabled=1):
        """Return a list of user objects
        If exclude_disabled is false, a user object is returned even if the user is
        disabled/meant to be acquired
        """
        if exclude_disabled:
            return map(self.getUser, self.getUserNames())
        else:
            return map(lambda name, self=self: self.getUser(name, 0), self.getUserNames(0))

    def getUser(self, name, exclude_disabled=1):
        """Return the named user object or None
        If exclude_disabled is false, the user object is returned even if the
        user is disabled/meant to be acquired
        """
        if name is None:
            # In remote user mode, if there is no REMOTE_USER variable, getUser
            # is called with name==None
            return None
        name = self._normalize(name)
        user = self.data.get(name, None)
        if user is None:
            name, ntdom, ntname = self._parse_ntusername(name)
            info = self._get_userinfo(ntname, ntdom)
            user = self.data.get(name, None)
            if user is None:
                if ntdom is None or info is None:
                    # non existent user
                    return None
                # the user has not be given roles yet
                # give a non persistent created-on-the-fly user object
                user = NTUser(name, None, self.default_roles, self.default_domains, ntdom, ntname,
                              info['full_name'], info['comment'], self.default_acquire,
                              ntuser=1)
        if exclude_disabled and self.isDisabled(user):
            return None
        return user

    def _addUser(self, name, password, confirm, roles, domains, REQUEST=None):
        name, ntdom, ntname = self._parse_ntusername(self._normalize(name), strict=1)
        if ntdom is not None:
            # a NT User
            if password == confirm == '':
                # make UserFolder don't complain about missing password
                password = confirm = 'x'
        return UserFolder._addUser(self, name, password, confirm, roles, domains, REQUEST)

    def _changeUser(self, name, password, confirm, roles, domains, REQUEST=None):
        name, ntdom, ntname = self._parse_ntusername(self._normalize(name), strict=1)
        if ntdom is not None:
            # a NT User
            if password == confirm == '':
                # make UserFolder don't complain about missing password
                # Zope >= 2.2 doesn't require password and confirm. Zope 2.1.6 does.
                password = confirm = 'x'
        return UserFolder._changeUser(self, name, password, confirm, roles, domains, REQUEST)

    def _doChangeUser(self, name, password=None, roles=None, domains=None,
                      full_name=None, comment=None, acquire=None):
        name, ntdom, ntname = self._parse_ntusername(self._normalize(name), strict=1)
        info = self._get_userinfo(ntname, ntdom)
        user = self.data.get(name, None)
        if user is None:
            # Physically create the user in the user folder
            if roles is None: roles = self.default_roles
            if domains is None: domains = self.default_domains
            if info is None:
                # This is not a NT User from the local domain. It does not acquire settings by default
                if acquire is None: acquire = 0
                if ntdom is None:
                    # Zope user
                    user = NTUser(name, password, roles, domains, ntdom, ntname,
                                  full_name or '', comment or '', acquire,
                                  ntuser=0)
                    self.data[name] = user
                else:
                    # NT User from another NT Domain
                    # XXX One day, find a way to ask for the controller of
                    # the domain, and ask it the full_name and comment
                    user = NTUser(name, None, roles, domains, ntdom, ntname,
                                  '', '', acquire,
                                  ntuser=1)
                    self.data[name] = user
            else:
                if acquire is None: acquire = self.default_acquire
                user = NTUser(name, None, roles, domains, ntdom, ntname,
                              info['full_name'], info['comment'], acquire,
                              ntuser=1)
                self.data[name] = user
        else:
            # user already exists in the NTUserFolder. Change it
            if user.ntuser:
                # NT User
                if info is not None:
                    # update full_name and comment if necessary
                    if user.full_name != info['full_name']: user.full_name = info['full_name']
                    if user.comment != info['comment']: user.comment = info['comment']
            else:
                # Zope user
                if password is not None: user.__ = password
                if full_name is not None and user.full_name != full_name: user.full_name = full_name
                if comment is not None and user.comment != comment: user.comment = comment
            # Common properties
            if roles is not None and user.roles != roles:
                user.roles = roles
            if domains is not None and user.domains != domains:
                user.domains = domains
            if acquire is not None and user.acquire != acquire:
                user.acquire = acquire
    _doAddUser = _doChangeUser

    def _doDelUsers(self, names):
        for name in names:
            name, ntdom, ntnam = self._parse_ntusername(self._normalize(name), strict=1)
            if self.data.has_key(name):
                del self.data[name]

    def _applyDefaultSettings(self, name):
        """restore default user settings.

        acquire, roles and domains are restored to the default values
        """
        name, ntdom, ntname = self._parse_ntusername(self._normalize(name), strict=1)
        info = self._get_userinfo(ntname, ntdom)
        user = self.data.get(name, None)
        if user is None:
            # user does not exist. it already get the defaults
            pass
        else:
            if user.ntuser and info is not None:
                # user from local domain
                # delete it to get back the defaults
                del self.data[name]
            else:
                # zope user or user from another NT Domain
                user.roles = self.default_roles
                user.domains = self.default_domains
                user.acquire = self.default_acquire
                self.data[name] = user

    _editPropertiesForm = HTMLFile('www/editPropertiesForm', globals(), __name__='manage_editPropertiesForm')
    def manage_editPropertiesForm(self, REQUEST=None, **kw):
        """Edit NTUserFolder properties"""
        local_ntdom = get_local_ntdom()
        if local_ntdom != self.local_ntdom:
            self.local_ntdom = local_ntdom
        kw['local_ntdom'] = local_ntdom
        kw['default_ntdom'] = self.default_ntdom
        kw['default_roles'] = self.default_roles
        kw['default_domains'] = self.default_domains
        return apply(self._editPropertiesForm, (self, REQUEST), kw)
    
    def manage_editProperties(self, ntdom, roles=[], domains=[], REQUEST=None):
        """Edit NTUserFolder properties:
        Change the default NT Domain
        """
        ntdom = lower(strip(ntdom))
        self.default_ntdom = ntdom or self.local_ntdom
        self.default_roles[:] = roles
        self.default_domains[:] = domains
        
        if REQUEST:
            message = 'NTUserFolder properties changed'
            return self.manage_editPropertiesForm(REQUEST, manage_tabs_message=message)

    manage_viewForm = HTMLFile('www/viewForm', globals(), __name__='manage_viewForm')
    
    manage_editUserForm = HTMLFile('www/editUserForm', globals(), __name__='manage_editUserForm')

    def manage_editUser(self, name, roles=None, domains=None,
                        full_name=None, comment=None, acquire=0,
                        apply_default_settings=0,
                        REQUEST=None):
        """Edit user properties"""
        if apply_default_settings:
            self._applyDefaultSettings(name)
        else:
            acquire = acquire and 1 or 0
            if roles is None: roles = []
            if domains is None: domains = []
            self._doChangeUser(name, roles=roles, domains=domains,
                               full_name=full_name, comment=comment, acquire=acquire)

        if REQUEST:
            message = 'User properties changed'
            return self.manage_viewForm(self, REQUEST, manage_tabs_message=message)

default__class_init__(NTUserFolder)

_addNTUserFolderForm = HTMLFile('www/addNTUserFolderForm', globals())
def manage_addNTUserFolderForm(self, REQUEST=None, **kw):
    """Add NT User Folder Form"""
    kw['local_ntdom'] = get_local_ntdom()
    kw['isTop'] = _isTop(self)
    kw['valid_roles'] = self.valid_roles()
    return apply(_addNTUserFolderForm, (self, self, REQUEST), kw)

def manage_addNTUserFolder(self, ntdom, acquire=0,
                           roles=None, domains=None,
                           replace_userfolder=0,
                           REQUEST=None):
    """Add NT User Folder"""
    # Install functions of OM that we need
    # Zope 2.1.6 does not have _getOb
    # Zope <= 2.3.x does not have _delObject
    for name in ('_getOb', '_delObject'):
        if not hasattr(self.__class__, name):
            setattr(self.__class__, name, Acquired)

    old_uf = self._getOb('acl_users', None)
    ntuf = NTUserFolder(ntdom, acquire, roles, domains, _isTop(self))
    if old_uf is None:
        initusers = ()
    else:
        if not replace_userfolder or \
           getattr(old_uf, '_isAjcNTUserFolder', 0):
            if REQUEST:
                return MessageDialog(
                    title = 'Item Exists',
                    message = 'This object already contains a User Folder',
                    action = '%s/manage_main' % REQUEST['URL1'])
            else:
                raise ValueError, "This object already contains a User Folder"
        else:
            # try to replace the old user folder after migrating the content
            initusers = old_uf.getUsers()
            self._delObject('acl_users')

    self._setObject('acl_users', ntuf)
    self.__allow_groups__ = ntuf
    ntuf = self._getOb('acl_users') # get the user folder in the right context (I hope)
    for user in initusers:
        name = user.getUserName()
        if ntuf.getUser(name) is None:
            # create the new user
            ntuf._doAddUser(name, user._getPassword(),
                          user.getRoles(), user.getDomains())
        else:
            # edit domains and roles
            ntuf._doChangeUser(name, None,
                               user.getRoles(), user.getDomains())

    if REQUEST:
        REQUEST['RESPONSE'].redirect(self.absolute_url() + '/manage_main?update_menu=1')

class NTUser(User):
    """NT User"""

    __ac_permissions__=(
        ('Manage users',
         ('isAcquiringSettings', 'isNTUser', 'authenticate'
          )),
        )

    def __init__(self, name, password, roles, domains, ntdom=None, ntname=None,
                 full_name='', comment='', acquire=0,
                 ntuser=0):
        User.__init__(self, name, password, roles, domains)
        
        self.ntdom = ntdom
        self.ntname = ntname
        self.full_name = full_name
        self.comment = comment
        self.acquire = acquire
        self.ntuser = ntuser

    getPrettyUserName__roles__ = None
    getPrettyUserName = User.getUserName # b/w compatibility

    getFullUserName__roles__ = None
    def getFullUserName(self):
        """Return the full name of the user"""
        return self.full_name

    getDescription__roles__ = None
    def getDescription(self):
        """Return description of the user"""
        return self.comment

    def isAcquiringSettings(self):
        """Return true if this user is acquiring settings"""
        return self.acquire

    def isNTUser(self):
        """Return true is this user is a NT User"""
        return self.ntuser
    
    def authenticate(self, password, request):
        """try to authenticate the user"""
        if self.ntuser:
            return (not self.domains or domainSpecMatch(self.domains, request)) and \
                   authenticate_user(self.ntname, self.ntdom, password)
        else:
            return User.authenticate(self, password, request)

default__class_init__(NTUser)
